package sun.beanbox;

/**
 * AppletGenerator.java
 *
 * Generates "Composed" Beans and Applets
 *
 * We will populate a directory with several useful files to create
 * and test a new bean.
 *
 * The generated JAR contains a class that explicitly knows what are its
 * top level beans, which are stored in a single serialized stream
 * together with their hookups.
 * This makes it easy to access the classloader for itself,
 * makes it easy to customize by adding properties, methods, etc.
 *
 */

import java.io.*;
import java.util.*;
import java.awt.Rectangle;
import java.awt.Frame;

public class AppletGenerator {
    
    public AppletGenerator(Frame ff,
			   BeanBox box,
			   File adir,
			   String aname,
			   String jname,
			   File tdir) {
        frame = ff;
	beanBox = box;
	appletName = aname;
	jarName = jname;
	appletDir = adir;
	tmpDir = tdir;
	initialize();
    }

    /**
     * Generate a new Applet from the contents of given BeanBox
     * appletName is the name of the Applet
     * appletDir is the directory where to put all the generated files
     * jarName is the name of the JAR where to package things.
     */

    public static void generate(Frame frame,
				BeanBox bb,
				MakeAppletDialog dialog,
				String appletDirName,
				String appletName,
				String jarName)
    {
	File tmpdir = new File(appletDirName, jarName+"_files");
	if (! tmpdir.exists()) {
	    tmpdir.mkdirs();
	}

	if (! tmpdir.isDirectory()) {
	    new ErrorDialog(frame, "Could not create directory "+tmpdir);
	    return;
	}

	if (checkForExternalizable(bb)) {
	    new ErrorDialog(frame, "The September 97 BDK cannot make an Applet\nwhen the BeanBox has an Externalizable component");
	    return;
	}

	File beandir = new File(appletDirName);

	AppletGenerator g = new AppletGenerator(frame, bb, beandir, appletName, jarName, tmpdir);

	try {
	    dialog.updateMessage("Copying Jars...");
	    g.copyJars(); 	// copy needed jar files
	    dialog.updateMessage("Writing and Compiling AppletClass...");
	    g.writeAppletClass(); // write and compile the applet class itself
	    dialog.updateMessage("Writing InitializeData...");
	    g.writeInitializeData(); // initialized data
	    dialog.updateMessage("Copying Hookups...");
	    g.copyHookups();	// copy the hookups (if any)
	    dialog.updateMessage("Writing JarFile...");
	    g.writeJarFile();	// create the Jar itself

	    dialog.updateMessage("Writing README file...");
	    g.writeReadme();	// readme documentation
	    dialog.updateMessage("Writing HTML file...");
	    g.writeHtml();	// a test case
	    dialog.updateMessage("Writing Makefiles...");
	    g.writeMakefiles();	// Makefiles for manual regeneration
	    dialog.updateMessage("Done");

        } catch (Exception  ex) {
	    new ErrorDialog(frame, 
			    ex.toString()+"\n"
			    +"while generating data for "+appletName);
	}
    }


    /**
     * Initialization
     */

    private void initialize() {
	initializeJars();
	initializeHookups();
	String bdkHome = System.getProperty("user.dir");
	supportJar = new File(new File(bdkHome, "lib"), "support.jar");
    }

    /* compute the jar dependencies */
    private void initializeJars() {
	longJars = new Hashtable();
	int count = beanBox.getComponentCount();
	for (int i=0; i<count; i++) {
	    Wrapper w = (Wrapper) beanBox.getComponent(i);
	    String jarName = JarInfo.getJarName(w.getBeanName());
	    String shortName =
		jarName.substring(jarName.lastIndexOf(File.separator)+1);
	    longJars.put(shortName, jarName);
	}
	String back[] = new String[longJars.size()];
	Enumeration e = longJars.keys();
	for (int i = 0; e.hasMoreElements(); i++) {
	    back[i] = (String) e.nextElement();
	}
	shortJars = back;
    }

    /* compute the hookup dependencies */
    private void initializeHookups() {
	Hashtable h = new Hashtable();
	int count = beanBox.getComponentCount();
	for (int i=0; i<count; i++) {
	    Wrapper w = (Wrapper) beanBox.getComponent(i);
	    Class[] k = w.getEventTargetClasses();
	    for (int j = 0; j < k.length; j++) {
		String s  = k[j].getName();
                if (! s.equals("sunw.beanbox.PropertyHookup")) {
		    h.put(s, s);
		}
	    }
	}
	hookups = new String[h.size()];
	Enumeration e = h.keys();
	for (int i = 0; e.hasMoreElements(); i++) {
	    hookups[i] = (String) e.nextElement();
	}
    }
  
    /**
     * Check if we have any Externalizable components
     *
     * It is a limitation of the September 97 BDK that it will not work properly
     * with Externalizable components.
     */

    static boolean checkForExternalizable(BeanBox bb) {
	int count = bb.getComponentCount();
	for (int i=0; i<count; i++) {
	    Wrapper w = (Wrapper) bb.getComponent(i);
	    Object b = w.getBean();
	    if (java.io.Externalizable.class.isInstance(b))
		return true;
	}
	return false;
    }

    /*
     * Generate a bean class
     */

    private void writeAppletClass() throws IOException {
	AppletClassGenerator g
	    = new AppletClassGenerator(beanBox);

	// dependent classes, as full names
	String jarNames[] = new String[longJars.size()];
	Enumeration e = longJars.elements();
	for (int i = 0; e.hasMoreElements(); i++) {
	    jarNames[i] = (String) e.nextElement();
	}
	boolean ok = g.generateClass(appletName, tmpDir, jarNames);
	if (!ok) {
	    throw new IOException("compilation failed!");
	}
    }

    /*
     * generate the serialized data that will be read later.
     * this is an serialied array,
     * first the string (classname of generated bean)
     * then serialized data for each of the components in the bean.
     */

    private void writeInitializeData()  throws IOException {
	File f = new File(tmpDir, appletName + "Data");
	ObjectOutputStream oos = null;
	try {
	    oos = new ObjectOutputStream(new FileOutputStream(f));
	} catch (Exception ex) {
	    System.err.println("could not create output stream for serialized data; "+ex);
	    System.exit(1);
	}
	int count = beanBox.getComponentCount();
	Object[] data = new Object[count+1];
	Rectangle[] boundsData = new Rectangle[count+1];

	data[0] = appletName;
	boundsData[0] = beanBox.getBounds();

	for (int i=0; i<count; i++) {
	    Wrapper w = (Wrapper) beanBox.getComponent(i);
	    Object bean = w.getBean();
	    data[i+1] = bean;
	    if (bean instanceof java.awt.Component) {
                Rectangle wr = w.getBounds();
		Rectangle br = ((java.awt.Component) bean).getBounds();
                boundsData[i+1] = new Rectangle(wr.x+br.x,
						wr.y+br.y,
						br.width,
						br.height);
	    } // else null
	}

	// Remove the Wrapper listening to the bean (if new event model)
	for (int i=0; i<count; i++) {
	    Wrapper w = (Wrapper) beanBox.getComponent(i);
	    w.listenForMice(false);
	} 

	oos.writeObject(data); // write the whole array as a single data.
	oos.writeObject(boundsData);
	
	// Rstore the Wrapper listening to the bean (if new event model)
	for (int i=0; i<count; i++) {
	    Wrapper w = (Wrapper) beanBox.getComponent(i);
	    w.listenForMice(true);
	} 

	oos.close();
    }

    /**
     * Copy here the hookup classes we need
     */
    private void copyHookups() throws IOException {
	String hookups[] = getHookups();
	for (int i=0; i<hookups.length; i++) {
	    String s = hookups[i].replace('.', File.separatorChar).concat(".class");
	    File src = new File(s);
	    File dst = new File(tmpDir, s);
	    copyFile(src, dst);
	}
    }

    /**
     * Copy here the Jars that this bean depends on
     */
    private void copyJars() throws IOException {
	// copy support jar
	copyFile(supportJar, new File(appletDir, "support.jar"));

	// copy other jars
	String jars[] = getShortJars();
	for (int i=0; i<jars.length; i++) {
	    File src = new File(getLongJar(jars[i]));
	    File dst = new File(appletDir, jars[i]);
	    copyFile(src, dst);
	}
    }

    /**
     * Copy a file across
     */
    private boolean copyFile(File src, File dst) throws IOException {
	if (! src.exists()) {
	    new ErrorDialog(frame, "Internal error: "+src+" does not exist");
	    return false;
	}
	File dstDir = new File(dst.getParent());
	if (! dstDir.exists()) {
	    dstDir.mkdirs();
	}
	if (! dstDir.isDirectory()) {
	    new ErrorDialog(frame, "Could not create directory "+dstDir);
	    return false;
	}

	byte buff[] = new byte[1024];
	int length = (int) src.length();

	InputStream is = new FileInputStream(src);
	OutputStream os = new FileOutputStream(dst);
	int nbytes;
	while ((nbytes = is.read(buff)) > 0) {
	    os.write(buff, 0, nbytes);
	}
	is.close();
	os.close();
	return true;
    }


    /**
     * Generate the JAR file.
     * This is equivalent to doing the corresponding make
     */

    private void writeJarFile() throws IOException {
	String[] hookups = getHookups();
	String[] files = new String[hookups.length+2];
	files[0] = appletName+".class";
	files[1] = appletName+"Data";
	for (int i=0; i<hookups.length; i++) {
	    files[i+2]
		= hookups[i].replace('.', File.separatorChar).concat(".class");
	}

	OutputStream out
	    = new FileOutputStream(new File(appletDir, jarName+".jar"));
	JarAccess.create(out, tmpDir, files[0], files);
    }

    /**
     * Generate Makefiles so users can do minor tweaks to the generated Applets
     */

    private final static int win32=0;
    private final static int unix=1;
    private final static int other=99;

    private void writeMakefiles() {
	writeMakefile(win32, "Makefile");
	writeMakefile(unix, "GNUmakefile");
    }

    /**
     * Write a makefile of a given type, and a given name
     */
    private void writeMakefile(int type, String mname) {
	File f;
	char fsep;		// file separator
	char psep;		// path separator
	String copy;		// the copy command

	switch (type) {
	case win32:
	    f = new File(tmpDir, mname);
	    fsep = '\\';
	    psep = ';';
	    copy = "copy";
	    break;
	case unix:
	    f = new File(tmpDir, mname);
	    fsep = '/';
	    psep = ':';
	    copy = "/bin/cp";
	    break;
	default:
	    return;
	}

	try {
	    IndentedStream p
	    	= new IndentedStream(new PrintWriter(new FileOutputStream(f)),
				     "	");
	    p.pp();

	    p.pp("BDKSUPPORTJAR= \\");
	    p.ip(supportJar.getPath());
	    p.op();

	    p.pp("JARS= \\");
	    p.i();
	    p.pp0("$(BDKSUPPORTJAR)");
	    String jars[] = getShortJars();
	    for (int i=0; i<jars.length; i++) {
		p.pn(" \\");
		p.pp0(".."+fsep+jars[i]);
	    }
	    p.pp();
	    p.op();
	    
	    p.pp0("JARPATH=.."+fsep+"support.jar");
	    for (int i=0; i<jars.length; i++) {
		p.pn0(psep+".."+fsep+jars[i]);
	    }
	    p.pp();
	    p.pp();

	    p.pp("JARFILE= \\");
	    p.ip(".."+fsep+jarName+".jar");
	    p.op();

	    String[] hookups = getHookups();
	    p.pp0("HOOKUPS=");
	    p.i();
	    for (int i=0; i<hookups.length; i++) {
		p.pn(" \\");
		p.pp0(hookups[i].replace('.', fsep).concat(".class"));
	    }
	    p.pp();
	    p.op();

	    p.pp("##");
	    p.pp();

	    p.pp("all:	$(JARFILE)");
	    p.pp();

	    p.pp("$(JARFILE): "+appletName+".class "+appletName+"Data $(HOOKUPS)");
	    if (type == unix) {
		p.ip("echo \"Name: "+appletName+".class\" >> manifest.tmp");
		p.pp("echo \"Java-Bean: True\" >> manifest.tmp");
		p.pp("jar cfm $(JARFILE) manifest.tmp "+appletName+".class "+appletName+"Data $(HOOKUPS)");
		p.pp("@/bin/rm manifest.tmp");
		p.op();
	    } else if (type == win32) {
		p.ip("jar cfm $(JARFILE) <<manifest.tmp "+appletName+".class "+appletName+"Data $(HOOKUPS)");
		p.op("Name: "+appletName+".class");
		p.pp("Java-Bean: True");
		p.pp("<<");
		p.pp();
	    }

	    p.pp(appletName+".class: "+appletName+".java $(JARS)");
	    if (type == unix) {
		p.ip("export CLASSPATH; CLASSPATH=."+psep+"$(JARPATH)"+psep+"$(CLASSPATH); \\");
	    } else if (type == win32) {
		p.ip("set CLASSPATH=."+psep+"$(JARPATH)"+psep+"%CLASSPATH%");
	    }
	    p.pp("javac "+appletName+".java");
	    p.op();

	    p.pp(".."+fsep+"support.jar: $(BDKSUPPORTJAR)");
	    p.ip(copy+" $(BDKSUPPORTJAR) .."+fsep+"support.jar");
	    p.op();

	    for (int i=0; i<jars.length; i++) {
		String sname = jars[i];
		String lname = getLongJar(sname);
		p.pp(".."+fsep+sname+": "+lname);
		p.ip(copy+" "+lname+" .."+fsep+sname);
		p.op();
	    }

	    p.close();
	} catch (Exception ex) {
	    status = "Trouble generating "+f.getName()+" file; ex: "+ex.toString();
	}
    }

    private void writeReadme() {
	File f = new File(appletDir, jarName+"_readme");
	try {
	    IndentedStream p
	    	= new IndentedStream(new PrintWriter(new FileOutputStream(f)),
				     "	");

	    p.pp("MakeApplet generates an Applet corresponding to a BeanBox layout.  The");
	    p.pp("data and classes needed for this Applet are packaged into a JAR file.");
	    p.pp("The Applet itself is a Bean and it can be read back into the BeanBox");
	    p.pp("if desired.");
	    p.pp("");
	    p.pp("The generated Applet shows one particular way of using serialization");
	    p.pp("to create a compound bean, other builders may use different");
	    p.pp("techniques.");
	    p.pp("");
	    p.pp("This directory contains the following files:");
	    p.pp("");
	    p.pp(jarName+".html");
	    p.pp("	A test html file for displaying the applet.  It requires");
	    p.pp("	a JDK1.1-compliant browser.  The file has specific comments for");
	    p.pp("	HotJava, IE4.0, and Navigator4.0.  The APPLET tag references");
	    p.pp("	the JAR files described next");
	    p.pp("");
	    p.pp(jarName+".jar");
	    p.pp("	This JAR file packages .class files and serialized data");
	    p.pp("	so it can be referenced by the HTML file.  This JAR file");
	    p.pp("	depends on the JAR files listed next.");
	    p.pp("");
	    p.pp("support.jar");
	    p.pp("	This JAR file contains a few support classes that are");
	    p.pp("	used the generated classes.");
	    p.pp("");
	    String[] jars = getShortJars();
	    for (int i=0; i<jars.length; i++) {
		p.pp(jars[i]);
	    }
	    p.pp("	These JAR file(s) describe the Beans that were instantiated");
	    p.pp("	in the BeanBox and packaged into MyApplet.");
	    p.pp("");
	    p.pp(jarName+"_files");
	    p.pp("	This directory contains source files, data, and Makefiles.");
	    p.pp("	It is possible to do small adjustments to the generated sources");
	    p.pp("	and create a new JAR file.");
	    p.pp("");

	    p.pp("");
	    p.pp("***************");
	    p.pp("");
	    p.pp("BROWSER SUPPORT");
	    p.pp("");
	    p.pp("The generated applet requires full JDK1.1 support.  The appletviewer");
	    p.pp("tool included in the JDK1.1 distributions can be used.  Alternatively");
	    p.pp("you can use the HotJava browser.");
	    p.pp("");
	    p.pp("The latest versions of Navigator and Communicator (4.0 and 4.01) do not");
	    p.pp("yet support enough of JDK1.1 for the generated applet to work.");
	    p.pp("The necessary support is expected later this year.");
	    p.pp("");
	    p.pp("Internet Explorer 4.0preview 2 provides significant JDK-1.1 support but");
	    p.pp("we are aware of two problems with it:");
	    p.pp("");
	    p.pp(" - it does not support JARs yet");
	    p.pp(" - it does not properly restore all information in a serialized component.");
	    p.pp("");
	    p.pp("The first problem can be addressed by modifying the APPLET tag in");
	    p.pp("myApplet.jar so it does not include the ARCHIVE= attribute.  Then");
	    p.pp("you will need to un-jar (using the \"jar xf\" command) all the JARs");
	    p.pp("listed in that attribute");
	    p.pp("");
	    p.pp("We know of no solution to the second problem.  In a typical example,");
	    p.pp("a deserialized button will not react to mouse events.");
	    p.pp("");

	    p.close();
	} catch (Exception ex) {
	    status = "Trouble generating "+f.getName()+" file; ex: "+ex.toString();
	}
    }

    private String[] getShortJars() {
	return shortJars;
    }

    private String getLongJar(String shortJar) {
	return (String) longJars.get(shortJar);
    }

    private String[] getHookups() {
	return hookups;
    }

    private void writeHtml() {
	File f = new File(appletDir, jarName+".html");
	try {
	    IndentedStream p
		= new IndentedStream(new PrintWriter(new FileOutputStream(f)));
	    p.pp("<html>");
	    p.pp("<head>");
	    p.pp("<title>Test page for "+appletName+" as an APPLET</Title>");
	    p.pp("</head>");
	    p.pp("<body>");
	    p.pp("<h1>Test for "+appletName+" as an APPLET</h1>");
	    p.pp("This is an example of the use of the generated");
	    p.pp(appletName+" applet.  Notice the Applet tag requires several");
	    p.pp("archives, one per JAR used in building the Applet");
	    p.pp("<p>");
	    p.pp("<applet");
	    p.ip("archive=\"./"+jarName+".jar,./support.jar");
	    String[] jars = getShortJars();
	    if (jars.length > 0) {
	    	p.i();
	    }
	    for (int i=0; i<jars.length; i++) {
	    	p.pp(",./"+jars[i]);
 	    }
	    p.op("\"");
	    p.pp("code=\""+appletName+"\"");
	    Rectangle bounds = beanBox.getBounds();
	    p.pp("width="+bounds.width);
	    p.pp("height="+bounds.height);
	    p.op(">");
	    p.pp("Trouble instantiating applet "+appletName+"!!");
	    p.pp("</applet>");
	    p.close();
	} catch (Exception ex) {
	    status = "Trouble generating "+f.getName()+" file; ex: "+ex.toString();
	}
    }

    // ===========
    // Private fields
    // ===========

    private File supportJar;	// Support jar

    private Frame frame;	// the frame for presenting messages
    private BeanBox beanBox;	// source for what we want to pack
    private String appletName;	// the name of the generated Applet
    private String jarName;	// the name of the jarFile (no .jar, no "/")
    private File tmpDir;	// where to generate the mess before packing it up
    private File appletDir;	// where to put the generated files
    private String[] shortJars; // short names for jars used in this beanbox
    private Hashtable longJars; // original names, indexed by short names
    private String[] hookups;	// hookups this beanbox depends on

    private String status;      // the status of this generation
}
